#include "transformation.h"

Transformation::Transformation(PNM* image) :
    QThread(0)
{
    parameters = new QHash<QString, QVariant>;
    this->image = image;
    this->supervisor = 0;
}

Transformation::Transformation(PNM* image, ImageViewer* iv) :
    QThread(0)
{
    parameters = new QHash<QString, QVariant>;
    this->image = image;
    this->supervisor = iv;

    if (iv)
    {
        connect(this, SIGNAL(started()), iv, SLOT(transformationStarted()));
        connect(this, SIGNAL(finished()), iv, SLOT(transformationFinished()));
        connect(this, SIGNAL(message(QString)), iv, SLOT(toolMessage(QString)));
        connect(this, SIGNAL(progress(int)), iv, SLOT(toolProgress(int)));
        connect(this, SIGNAL(resultReady(PNM*)), iv, SLOT(updateImage(PNM*)));
        connect(this, SIGNAL(finished()), iv, SLOT(toolFinished()));
    }
}

Transformation::~Transformation()
{
    delete parameters;
    if (supervisor)
    {
        disconnect(this, SIGNAL(started()));
        disconnect(this, SIGNAL(message(QString)));
        disconnect(this, SIGNAL(progress(int)));
        disconnect(this, SIGNAL(resultReady(PNM*)));
        disconnect(this, SIGNAL(finished()));
    }
}

/** Sets a parameter param of the transformation
* with the value.
* If the parameter was already set it is overridden
* Returns the pointer to the Transformation object so it can be used as a factory.
*/
Transformation* Transformation::setParameter(QString param, QVariant value)
{
    parameters->insert(param, value);
    return this;
}

/** Returns the value of the given param.
* If the param isn't set returns a default QVariant()
* which is an invalid variant
* @see QVariant::QVariant()
*/
QVariant Transformation::getParameter(QString param)
{
    return parameters->value(param, QVariant());
}

void Transformation::run()
{
    PNM* image = this->transform();
    emit resultReady(image);
}

/**
* /virtual, abstract!/
* Virtual method that do some transformations (based on parameter Hash)
* in the Transformation class it returns a pointer to
* the loaded image or a new null PNM image
*/
PNM* Transformation::transform()
{
    if (image)
        return image;
    else
        return new PNM();
}

/** Returns a pointer to the image stored in the class */
PNM* Transformation::getImage()
{
    return image;
}

QRgb Transformation::getPixel(int x, int y, Mode mode)
{
    switch (mode) {
    case CyclicEdge:    return getPixelCyclic(x,y);
    case NullEdge:      return getPixelNull(x,y);
    case RepeatEdge:    return getPixelRepeat(x,y);
    default:            return image->pixel(x,y);
    }
}

QRgb Transformation::getPixelCyclic(int x, int y) {
    int width = image->size().width();
    int height = image->size().height();
    return image->pixel(x % width,y % height);
}

QRgb Transformation::getPixelNull(int x, int y)
{
    const QRgb blackPixel = qRgb(0, 0, 0);

    int width = image->size().width();
    int height = image->size().height();
    if (x >= width || x < 0 || y >= height || y < 0)
        return blackPixel;

    return image->pixel(x,y);
}

QRgb Transformation::getPixelRepeat(int x, int y)
{
    int width = image->size().width();
    int height = image->size().height();

    if (x >= width) x = width - 1;
    if (x < 0) x = 0;

    if (y >= height) y = height - 1;
    if (y < 0) y = 0;
    return image->pixel(x,y);
}

math::matrix<float> Transformation::getWindow(int x, int y, int size,
                                              Channel channel = LChannel,
                                              Mode mode = RepeatEdge)
{
    math::matrix<float> window(size,size);

    for (int i=0; i<size; ++i) {
        for (int j=0; j<size; ++j) {

            QRgb pixel = getPixel(x - int(size/2) + i, y - int(size/2) + j, mode);

            int pixelVal;

            if (channel == RChannel) pixelVal = qRed(pixel);
            if (channel == GChannel) pixelVal = qGreen(pixel);
            if (channel == BChannel) pixelVal = qBlue(pixel);
            if (channel == LChannel) pixelVal = qGray(pixel);


            window[i][j] = pixelVal;
        }
    }

    return window;
}

ImageViewer* Transformation::getSupervisor()
{
    return supervisor;
}
